// Библиотеки для датчиков
#include <Adafruit_VL6180X.h>
#include <Adafruit_VL53L0X.h>

// Объявление адресов датчиков в шине I2C
#define LOX0_ADDRESS 0x29
#define LOX1_ADDRESS 0x30
#define LOX2_ADDRESS 0x31

VL53L0X_RangingMeasurementData_t measure1;
// Объявление пинов датчиков
#define SHT_LOX0 A0
#define SHT_LOX1 7
#define SHT_LOX2 12
#define GPIO_LOX1 4
#define GPIO_LOX2 4

#define TIMING_PIN 13

// Создание объектов для датчиков
Adafruit_VL53L0X lox0 = Adafruit_VL53L0X();
Adafruit_VL6180X lox1 = Adafruit_VL6180X();
Adafruit_VL6180X lox2 = Adafruit_VL6180X();

typedef enum {RUN_MODE_DEFAULT, RUN_MODE_TIMED, RUN_MODE_ASYNC, RUN_MODE_GPIO, RUN_MODE_CONT} runmode_t;

runmode_t run_mode = RUN_MODE_DEFAULT;
uint8_t show_command_list = 1;

// Объявление пинов для моторов
const int PinA1 = 5;  // (ШИМ) вывод 5 соединен с выводом А1
const int PinA2 = 6;  // (ШИМ) вывод 6 соединен с выводом А2
const int PinB1 = 10; // (ШИМ) вывод 10 соединен с выводом (pin) B1
const int PinB2 = 11; //(ШИМ) вывод 11 соединен с выводом (pin) B2
byte speed = 250;
int encoderPin1 = 3;
int encoderPin2 = 2;
volatile int lastEncoded = 0;
volatile long encoderValue = 0;
// Дополнительные переменные для датчиков расстояния
Adafruit_VL6180X *sensors[] = {&lox1, &lox2};
const uint8_t COUNT_SENSORS = sizeof(sensors) / sizeof(sensors[0]);
const int sensor_gpios[COUNT_SENSORS] = {GPIO_LOX1, GPIO_LOX2};
uint8_t  tempRange;
uint8_t  sensor_ranges[3];
uint8_t  sensor_status[COUNT_SENSORS];
const uint16_t  ALL_SENSORS_PENDING = ((1 << COUNT_SENSORS) - 1);
uint16_t  sensors_pending = ALL_SENSORS_PENDING;
uint32_t  sensor_last_cycle_time;

// Переменные для позиционирования 
int x = 0; 
int y = 0;
int nap = 1;

void setID() {
  // Сброс всех датчиков
  digitalWrite(SHT_LOX0, LOW);
  digitalWrite(SHT_LOX1, LOW);
  digitalWrite(SHT_LOX2, LOW);

  delay(10);

  // Включение всех датчиков
  digitalWrite(SHT_LOX0, HIGH);
  digitalWrite(SHT_LOX1, HIGH);
  digitalWrite(SHT_LOX2, HIGH);

  delay(10);

  // Активация LOX1 и сброс LOX2
  digitalWrite(SHT_LOX0, LOW);
  digitalWrite(SHT_LOX1, HIGH);
  digitalWrite(SHT_LOX2, LOW);

  // Инициализация LOX1
  if (!lox1.begin()) {
    Serial.println(F("Не удалось инициализировать первый VL6180X"));
    while (1);
  }
  lox1.setAddress(LOX1_ADDRESS);
  delay(10);

  // Активация LOX2
  digitalWrite(SHT_LOX2, HIGH);
  delay(10);

  // Инициализация LOX2
  if (!lox2.begin()) {
    Serial.println(F("Не удалось инициализировать второй VL6180X"));
    while (1);
  }
  lox2.setAddress(LOX2_ADDRESS);
  delay(10);

  // Активация LOX0
  digitalWrite(SHT_LOX0, HIGH);
  delay(10);

  // Инициализация LOX0
  if (!lox0.begin()) {
    Serial.println(F("Не удалось инициализировать третий VL6180X"));
    while (1);
  }
  lox0.setAddress(LOX0_ADDRESS);
  delay(10);
}

// Проверка 1 сенсора
uint8_t readSensor(Adafruit_VL6180X &vl) {
  float lux = vl.readLux(VL6180X_ALS_GAIN_5);
  uint8_t range = vl.readRange();
  uint8_t status = vl.readRangeStatus();

  if (status == VL6180X_ERROR_NONE) {
    return range;
  } else {
    return 255;
  }
}
// Проверка всех сенсеров
void read_sensors() {
  if(readSensor(lox1) < 150) {
    sensor_ranges[2] = 0;
  } else {
    sensor_ranges[2] = 1;
  }

  if(readSensor(lox2) < 150) {
    sensor_ranges[0] = 0;
  } else {
    sensor_ranges[0] = 1;
  }

  lox0.rangingTest(&measure1, false);
  

  if(measure1.RangeStatus != 4) {  // если не выходит за пределы диапазона
    if(measure1.RangeMilliMeter < 150) {
      sensor_ranges[1] = 0;
    } else {
      sensor_ranges[1] = 1;
    }
  } else {
    sensor_ranges[1] = 1;
  }
}

// Поворот направо
void right() {
  analogWrite(PinA1, speed);
  analogWrite(PinA2, 0);
  analogWrite(PinB1, speed);
  analogWrite(PinB2, 0);
}

// Поворот налево
void left() {
  analogWrite(PinA1, 0);
  analogWrite(PinA2, speed);
  analogWrite(PinB1, 0);
  analogWrite(PinB2, speed);
}

// Движение назад 
void backward() {
  analogWrite(PinA1, speed);
  analogWrite(PinA2, 0);
  analogWrite(PinB1, 0);
  analogWrite(PinB2, speed);
}

// Движение вперёд
void forward() {
  analogWrite(PinA1, 0);
  analogWrite(PinA2, speed);
  analogWrite(PinB1, speed);
  analogWrite(PinB2, 0);
}

// Остановка
void STOP() { 
  analogWrite(PinA1, 0);
  analogWrite(PinA2, 0);
  analogWrite(PinB1, 0);
  analogWrite(PinB2, 0);
  delay(200);
}

// поворот налево на 90 градусов
void left_90() {
  while(encoderValue < 2208) {
    left();
  }
  right();
  delay(50);
  encoderValue = 0;
  STOP();
  nap -= 1;
  if(nap <= 0) nap += 4;
}

// поворот направо на 90 градусов
void right_90() {
  while(encoderValue > -2168) {
    right();
  }
  left();
  delay(50);
  encoderValue = 0;
  STOP();
  nap += 1;
  if(nap >= 5) nap -= 4;
}

// поворот налево на 180 градусов
void left_180() {
  while(encoderValue < 2205 * 2) {
    left();
  }
  right();
  delay(50);
  encoderValue = 0;
  STOP();
  nap -= 2;
  if(nap <= 0) nap += 4;
}

// Проезд вперёд на 1 клетку
void vper() {
  while(encoderValue < 5134) {
    forward();
  }
  encoderValue = 0;
  if(nap == 1) y += 1;
  if(nap == 2) x += 1;
  if(nap == 3) y -= 1;
  if(nap == 4) x -= 1;
}

// Проезд вперёд на 1 клетку и остановка
void vper_stop() {
  while(encoderValue < 5434) {
    forward();
  }
  backward();
  delay(50);
  encoderValue = 0;
  STOP();
  if(nap == 1) y += 1;
  if(nap == 2) x += 1;
  if(nap == 3) y -= 1;
  if(nap == 4) x -= 1;
}

// Обнавление значений энкодоров
void updateEncoder() {
  int MSB = digitalRead(encoderPin1); // MSB = старший значащий бит
  int LSB = digitalRead(encoderPin2); // LSB = младший значащий бит

  int encoded = (MSB << 1) | LSB; // преобразование значений двух пинов в одно число
  int sum  = (lastEncoded << 2) | encoded; // добавление его к предыдущему закодированному значению

  if(sum == 0b1101 || sum == 0b0100 || sum == 0b0010 || sum == 0b1011) encoderValue--;
  if(sum == 0b1110 || sum == 0b0111 || sum == 0b0001 || sum == 0b1000) encoderValue++;

  lastEncoded = encoded; // сохранить это значение для следующего раза
}

// колличество возможных направлений двежения
int vos_nap() {
  int res = 0;
  for(int i = 0; i < 3; i++) {
    res += sensor_ranges[i];
  }
  return res;
}

void setup() {
  
//===============================================================
// Настройка пинов
//===============================================================

  Serial.begin(9600);

  // ждать, пока последовательный порт не откроется для устройств с родным USB
  while (!Serial) {
    delay(1);
  }

  pinMode(SHT_LOX1, OUTPUT);
  pinMode(SHT_LOX2, OUTPUT);
  pinMode(A0, OUTPUT);
  digitalWrite(A0, 0);
  // Включить пин синхронизации, чтобы было легко видеть, когда проходит передача данных
  pinMode(TIMING_PIN, OUTPUT);

#ifdef GPIO_LOX1
  // Если определены GPIO-пины, включить их как PULL UP
  pinMode(GPIO_LOX1, INPUT_PULLUP);
  pinMode(GPIO_LOX2, INPUT_PULLUP);
#endif

  Serial.println("Инициализация пинов выключения...");

  digitalWrite(SHT_LOX1, LOW);
  digitalWrite(SHT_LOX2, LOW);

  digitalWrite(TIMING_PIN, LOW);
  Serial.println("Все в режиме сброса... (пины низкие)");

  Serial.println("Запуск...");
  setID();
  pinMode(PinA1, OUTPUT); // установить контакты как выход
  pinMode(PinA2, OUTPUT);
  pinMode(PinB1, OUTPUT);
  pinMode(PinB2, OUTPUT);
  pinMode(encoderPin1, INPUT_PULLUP); 
  pinMode(encoderPin2, INPUT_PULLUP);

  digitalWrite(encoderPin1, HIGH); // включить подтягивающий резистор
  digitalWrite(encoderPin2, HIGH); // включить подтягивающий резистор
  

  // вызвать updateEncoder() при любом изменении high/low
  // на прерываниях 0 (пин 2) или прерываниях 1 (пин 3)
  attachInterrupt(0, updateEncoder, CHANGE); 
  attachInterrupt(1, updateEncoder, CHANGE);

  int fin_alg[30] = {};
  int dop_alg[20] = {};

  int fi = 0;
  int di = 0;
  
  bool flag = 0;
  read_sensors();
  while(readSensor(lox2) > 20) {
    delay(10);
  }

  //===============================================================
  // Основной цикл
  //===============================================================

  while(true) {
    delay(500);
    read_sensors();
    if(vos_nap() >= 2) {
      if(flag) {
        bool fls = 0;
        for(int i = 0; i < 3; i++) {
          if(sensor_ranges[i]) {
            if(fls == 0) {
              fls = 1;
            } else {
              if(i == 0) {
                left_90();
                vper_stop();
                dop_alg[di] = 1;
                di++;
                dop_alg[di] = 2;
                di++;
              } else if(i == 1) {
                vper_stop();
                dop_alg[di] = 2;
                di++;
              } else {
                right_90();
                vper_stop();
                dop_alg[di] = 3;
                di++;
                dop_alg[di] = 2;
                di++;
              }
              break;
            }
          }
        }
      } else {
        for(int i = 0; i < di; i++) {
          fin_alg[fi] = dop_alg[i];
          fi++;
        }
        di = 0;
        for(int i = 0; i < 3; i++) {
          if(sensor_ranges[i]) {
            if(i == 0) {
              left_90();
              vper_stop();
              dop_alg[di] = 1;
              di++;
              dop_alg[di] = 2;
              di++;
            } else if(i == 1) {
              vper_stop();
              dop_alg[di] = 2;
              di++;
            } else {
              right_90();
              vper_stop();
              dop_alg[di] = 3;
              di++;
              dop_alg[di] = 2;
              di++;
            }
            break;
          }
        }
      }
      flag = 0;
    } else if(vos_nap() == 1) {
      flag = 0;
      if(sensor_ranges[0]) {
        left_90();
        vper_stop();
        dop_alg[di] = 1;
        di++;
        dop_alg[di] = 2;
        di++;
      } else if(sensor_ranges[1]) {
        vper_stop();
        dop_alg[di] = 2;
        di++;
      } else {
        right_90();
        vper_stop();
        dop_alg[di] = 3;
        di++;
        dop_alg[di] = 2;
        di++;
      }
    } else {
      if(x == 2 && y == 2) {
        for(int i = 0; i < di; i++) {
          fin_alg[fi] = dop_alg[i];
          fi++;
        }
        break;
      } else {
        left_180();
        di--;
        for(; di >= 0;) {
          if(dop_alg[di] == 3) left_90();
          else if(dop_alg[di] == 1) right_90();
          else vper_stop();
          di--;
        }
        left_180();
        di = 0;
        flag = 1;
      }
    }
    
    for(int i = 0; i < fi; i++) {
      Serial.print(fin_alg[i]);
      Serial.print(" ");
    }
    Serial.println(fi);
  }
  for(int i = 0; i < fi; i++) {
    Serial.print(fin_alg[i]);
    Serial.print(" ");
  }
  fin_alg[fi] = 0;
  while(readSensor(lox2) > 20) {
    delay(10);
  }
  for(int i = 0; i < fi; i++) {
    if(fin_alg[i] == 3) {
      right_90();
    } else if(fin_alg[i] == 1) {
      left_90();
    } else {
      if(fin_alg[i + 1] == 2) {
        vper();
      } else {
        vper_stop();
      }
    }
  }
}

void loop() {
  delay(1);
}